#!/usr/local/bin/ruby
### skkserv.rb                                   -*- Mode: Ruby -*-

## Copyright (C) 1997-2000  Shugo Maeda

## Author:	Shugo Maeda <shugo@aianet.ne.jp>
## Version:	2.3.2

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 2, or (at
## your option) any later version.

## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with GNU Emacs; see the file COPYING.  If not, write to the
## Free Software Foundation, Inc., 59 Temple Place - Suite 330,
## Boston, MA 02111-1307, USA.

### Code:

require "thread"
require "socket"
require "getopts"

getopts("d", "p:", "sjis", "nocache", "cachedir:", "help", "version")

class SKKServer
  VERSION_STRING = "rskkserv-2.3.2 "
  
  DEFAULT_DICFILE = "/usr/local/share/skk/SKK-JISYO.L"
  DEFAULT_CACHEDIR = "/usr/local/share/skk"
  
  MAX_CLIENTS = 32
  
  CLIENT_END = ?0
  CLIENT_REQUEST = ?1
  CLIENT_VERSION = ?2
  CLIENT_HOST = ?3
  
  SERVER_ERROR = ?0
  SERVER_FOUND = ?1
  SERVER_NOT_FOUND = ?4
  SERVER_FULL = ?9
  
  BUFSIZ = 1024
  
  def SKKServer.main
    if $OPT_help
      usage
      exit(0)
    end
    if $OPT_version
      print("SKK SERVER ", VERSION_STRING, "\n")
      exit(0)
    end
    
    dic = ARGV.shift || DEFAULT_DICFILE
    
    begin
      skkserv = new(dic)
      unless $OPT_d
	exit! if fork
	Process::setsid
	exit! if fork
	Dir::chdir("/")
	File::umask(022)
	STDIN.reopen(open("/dev/null", "r"))
	console = open("/dev/console", "w")
	STDOUT.reopen(console)
	STDERR.reopen(console)
      end
      skkserv.mainloop
    rescue Errno::EADDRINUSE
      $stderr.print("rskkserv is already running.\n")
      exit(1)
    rescue
      $stderr.printf("%s: %s\n", $0, $!)
      exit(1)
    end
  end
  
  def SKKServer.usage
    $stderr.printf(<<USAGE, $0)
Usage: %s [OPTION] skk-dictionary-file

  -d                    debug mode
  -p port               open server using specified port
  --sjis                use Shift JIS dictionary
  --nocache             don't use cache
  --cachedir dir        cache file directory
  --help                display this help and exit
  --version             output version information and exit
USAGE
#'
  end
  
  def initialize(dic)
    hostname = Socket.gethostname
    @host = hostname + ":" + TCPsocket.getaddress(hostname) + ": "
    printf("server is on %s\n", @host) if $OPT_d
    @dictionary = SKKDictionary.new(dic, $OPT_cachedir || DEFAULT_CACHEDIR)
    service = $OPT_p || "skkserv"
    begin
      @server = TCPserver.open(service)
    rescue
      if service == "skkserv"
	service = 1178
	retry
      else
	raise
      end
    end
    if $OPT_d
      addr = @server.addr
      addr.shift
      printf("port number is %d\n", addr.join(":"))
    end
    begin
      print("loading tcpwrap ...\n") if $OPT_d
      require "tcpwrap"
      @server = TCPWrapper.new(File.basename($0), @server, true)
      print("done\n") if $OPT_d
    rescue
      print("failed\n") if $OPT_d
    end
    @nclients = 0
  end
  
  def mainloop
    accept_clients do |s|
      while cmdbuf = s.sysread(BUFSIZ)
	case cmdbuf[0]
	when CLIENT_END
	  printf("message from client %s: END\n", s) if $OPT_d
	  break
	when CLIENT_REQUEST
	  printf("message from client %s: WORD\n", s) if $OPT_d
	  kana = cmdbuf[1 .. cmdbuf.index(? )]
	  ret = ""
	  if kanji = @dictionary.search(kana)
	    ret.concat(SERVER_FOUND)
	    ret.concat(kanji)
	  else
	    ret.concat(SERVER_NOT_FOUND)
	    ret.concat(cmdbuf[1 .. -1])
	  end
	  printf("send: %s\n", ret) if $OPT_d
	  s.write(ret)
	when CLIENT_VERSION
	  printf("message from client %s: VERSION\n", s) if $OPT_d
	  printf("send: %s\n", VERSION_STRING) if $OPT_d
	  s.write(VERSION_STRING)
	when CLIENT_HOST
	  printf("message from client %s: HOST\n", s) if $OPT_d
	  printf("send: %s\n", @host) if $OPT_d
	  s.write(@host)
	else
	  printf("message from client %s: UNKNOWN: %d/\"%s\"\n",
		 s, command[0], command) if $OPT_d
	end
      end
    end
  end
  
  def accept_clients
    loop do
      s = @server.accept
      print(s, " is accepted.\n") if $OPT_d
      if @nclients >= MAX_CLIENTS
	begin
	  s.putc(SERVER_FULL)
	  printf("Connection rejected: too many clients.\n", @nclients) if $OPT_d
	ensure
	  s.shutdown
	  s.close
	  next
	end
      end
      @nclients += 1
      printf("There are %d clients.\n", @nclients) if $OPT_d
      Thread.start do
	begin
	  yield(s)
	ensure
	  s.shutdown
	  s.close
	  print(s, " is gone.\n") if $OPT_d
	  @nclients -= 1
	  printf("There are %d clients.\n", @nclients) if $OPT_d
	end
      end
    end
  end
end

class SKKDictionary
  include FileTest
  
  OKURI_ARI_LABEL = ";; okuri-ari entries.\n"
  OKURI_NASI_LABEL = ";; okuri-nasi entries.\n"
  
  if $OPT_sjis
    OKURI_ARI_REGEXP = /[\x81-\x9f\xe0-\xef][\x40-\x7e\x80-\xfc].*[a-z] $/n
  else
    OKURI_ARI_REGEXP = /[\xa1-\xfe][\xa1-\xfe].*[a-z] $/n
  end
  
  def initialize(path, cachedir = File.dirname(path))
    @path = path
    @file = open(@path)
    
    @cache_a = File.expand_path(File.basename(@path) + ".a", cachedir)
    @cache_n = File.expand_path(File.basename(@path) + ".n", cachedir)
    
    if !$OPT_nocache &&
	(exist?(@cache_a) &&  File.ctime(@path) < File.ctime(@cache_a)) &&
	(exist?(@cache_n) &&  File.ctime(@path) < File.ctime(@cache_n))
      @okuri_ari_table = read_cache(@cache_a)
      @okuri_nasi_table = read_cache(@cache_n)
      if @okuri_ari_table.nil? || @okuri_nasi_table.nil?
	make_table
      end
    else
      make_table
    end
    @mutex = Mutex.new
  end
  
  def make_table
    print("Making index table for dictionary ...\n") if $OPT_d
      
    @okuri_ari_table = []
    @okuri_nasi_table = []
    for line in @file
      if line == OKURI_ARI_LABEL
	@okuri_ari_table.push(@file.pos)
	break
      end
    end
    for line in @file
      if line == OKURI_NASI_LABEL
	@okuri_nasi_table.push(@file.pos)
	break
      else
	@okuri_ari_table.push(@file.pos)
      end
    end
    @okuri_ari_table.pop
    @okuri_ari_table.reverse!
    for line in @file
      @okuri_nasi_table.push(@file.pos)
    end
    @okuri_nasi_table.pop
    
    unless $OPT_nocache
      write_cache(@okuri_ari_table, @cache_a)
      write_cache(@okuri_nasi_table, @cache_n)
    end
  end
  
  def write_cache(table, file)
    printf("Writing cache file %s ...\n", file) if $OPT_d
    open(file, "w") do |f|
      f.write(table.pack("i*"))
    end
  end
  
  def read_cache(file)
    printf("Reading cache file %s ...\n", file) if $OPT_d
    open(file) do |f|
      return f.read.unpack("i*")
    end
  end
  
  def search(kana)
    @mutex.synchronize do
      printf("target: %s\n", kana) if $OPT_d
      
      if kana =~ OKURI_ARI_REGEXP
	table = @okuri_ari_table
	print("okuri: ARI\n") if $OPT_d
      else
	table = @okuri_nasi_table
	print("okuri: NASI\n") if $OPT_d
      end
      
      left = 0
      right = table.length - 1
      while left <= right
	pos = (left + right) / 2
	@file.seek(table[pos], 0)
	buff = @file.readline
	print(buff) if $OPT_d
	case buff[0, kana.length] <=> kana
	when 0
	  return buff[kana.length .. -1]
	when 1
	  right = pos - 1
	when -1  
	  left = pos + 1
	end
      end
      
      return nil
    end
  end
end

SKKServer.main

### skkserv.rb ends here
